
%if 0

RxANSI: Combined ANSI and LESS driver. AMIS v3.6 compliant.
Public Domain by Henrik Haftmann and Christian Masloch

RxANSI.ASM - Resident code of RxANSI

Read RxANSI.DOC for more information.

%endif

%if 0

Supported Int2D functions:

AMIS - Installation check
INP:	al = 00h
OUT:	al = 0FFh
	cx = Private version number (currently 010Ah)
	dx:di-> signature: "cm      ","RxANSI  ",0 (don't depend on the description)

AMIS - Get private entry point - NOP: no private entry point
INP:	al = 01h
OUT:	al = 00h

AMIS - Uninstall - NOP: no resident uninstaller or NOP: can't uninstall
INP:	al = 02h
OUT:	If installed from command line,
	 al = 03h
	 bx = memory block of resident RxANSI (cs)
	If internal (pre-installed),
	 al = 01h
	(If with resident uninstaller,
	 al = 01h if unsuccessful
	 al = 0FFh if successful)

AMIS - Request pop-up - NOP: no pop-up
INP:	al = 03h
OUT:	al = 00h

AMIS - Determine chained interrupts
INP:	al = 04h
OUT:	al = 04h
	dx:bx-> interrupt hook list (Int29, Int2D always. Optionally Int2F.)

AMIS - Get hotkeys - NOP: no hotkeys
INP:	al = 05h
OUT:	al = 00h

AMIS - Get device driver information - NOP: no device
INP:	al = 06h
OUT:	al = 00h

AMIS - Reserved for AMIS
INP:	al = 07h..0Fh
OUT:	al = 00h

RxANSI - Get compatible version
INP:	al = 10h
OUT:	al = 0FFh
	cx = Compatible version number (currently 010Ah)

RxANSI - Get ctrl0
INP:	al = 20h
OUT:	al = size of returned control data (not 0)
	dx:bx-> control data
	LESS bit switched off if non-LESS version

RxANSI - Reserved for RxANSI
INP:	al = 11h..1Fh, 21h..FFh
OUT:	al = 00h

%endif

%if 0

To assemble RxANSI internal (pre-installed), do this:

- Make sure __AMIS_NUMBER__ is a valid and yet unallocated AMIS multiplex number.
  __AMIS_NUMBER__ should be initialized to 00h and increases for each installed
  AMIS interface. (Abort assembling if __AMIS_NUMBER__ is >0FFh before installing
  the next AMIS interface.)
- Include RXANSI.ASM into a source part that will be hold resident

To use RxANSI as the internal Int29 handler, RxBIO has to perform the following:

- Set Int29 to rxansiseg:rxansi_i29
- Save Int2D to [rxansiseg:rxansi_i2d.next]
- Set Int2D to rxansiseg:rxansi_i2d
- If Int2F is to be installed, do the same for ..._i2f
- Of course hold the memory utilised by RXANSI resident

That's it.

Considerations:	The Int2D and (especially) Int2F handlers could be more integrated into RxBIO.
		If functions of the resident RxANSI are in RxBIO too, they should be merged.

%endif

%include "CMMACROS.MAC"
%include "AMIS.MAC"

%ifnnum RXANSI_VERP
	verdef RXANSI,C, 1,10		; Version number of last compatible
	verdef RXANSI,P, 1,10		; Private version
%elif RXANSI_VERP != ver(1,10)		; if already defined and doesn't match current
 %error "RxANSI version doesn't match"
%endif

	numdef RXANSI, 1, 1
	numdef RXBIO
	numdef RXANSI_RESUNINST		; include resident uninstaller
	numdef RXANSI_NOLESS		; don't parse LESS sequences
	numdef RXANSI_NOI2F		; Don't hook interrupt 2Fh

rxansi_handler:

				; AMIS stuff
rxansi_amissig:
.ven:	fill 8,32,db "cm"		; vendor
.prod:	fill 8,32,db "RxANSI"		; product
.desc:	asciz				; save memory by omitting the description

rxansi_amisintr:
%ifn _RXBIO				; if we're inside RxBIO, we are the _last_ handler
.i29:	db 29h
	dw rxansi_i29
%endif
%ifn _RXANSI_NOI2F
.i2f:	db 2Fh
	dw rxansi_i2f
%endif
.i2d:	db 2Dh				; the only entry that's always here
	dw rxansi_i2d


				; RxANSI data
rxansi_ctrl0:	db 1100_0000b		; Control byte
					; Bit 7: Process ANSI sequences
					;     6: Process LESS sequences
					;     5: Last character was a digit
					;     4: former reversed
					;     3: former hidden
					;     2-0: former foreground color if hidden
rxansi_last:	dw 0			; Last two characters
rxansi_argn:	db 0			; Number of arguments
rxansi_argv:	times 5 db 0		; Argument values
rxansi_curv:	dw 0			; Saves cursor position
;rxansi_ctrl1:
rxansi_attrib:	db 7			; Character color

%if _RXANSI_RESUNINST
			; Code called by the resident uninstallation.
		; INP:	ds:bx -> AMIS interrupt list
		;	dx = code called for each interrupt (with ds:si -> interrupt entry)
		; OUT:	CY if any interrupt failed,
		;	 al = interrupt that failed to uninstall
		;	NC if all interrupts successful uninstalled,
		;	 al = 2Dh
		; CHG:	si, ax, bx, what called code (dx) changes
		;
		; Called code (dx) has to preserve dx, bx, cx, al.
rxansi_loopamisintr:
.loop:
	mov al, [bx]
	mov si, [bx +1]

	call dx
	jc .error			; error -->
	add bx, byte 3
	cmp al, 2Dh
	jne .loop			; do until 2Dh done -->
					; else ZR, NC
.error:
	retn

		; INP:	al = interrupt number
		;	ds:si-> interrupt entry
		; OUT:	CY if unhooking failed
		;	NC if unhooking successful
		; CHG:	ah, es, di
rxansi_unhookinterrupt:
			; unhookinterruptsim (below) only checks if it's possible to unhook this interrupt.
			; This function really unhooks the interrupt if possible.
			;
			; This is to cover the situation when some of the hooked interrupts can unhook,
			; but some can't. If the uninstaller would start to unhook the interrupts and then
			; catch the interrupt that can't be unhooked the user would end with a dead TSR
			; that's uninstalled halfway. Very bad.
			;
			; "Simulating" the unhooking first and checking if all Interrupts can unhook
			; normally will not return such a state.
	call rxansi_unhookinterruptsim
	jc .ret				; bad. -->
	jz .easy
.hard:
				; "hard" case: unhookinterruptsim has however already done the work,
				; so the hard case is here indeed easier than the easy case.
	cmpsw				; skip 10EBh or 0EA90h
	movsw				; copies our stored pointer into the other's entry
	movsw				; copies our stored segment into the other's entry
	clc
	retn
.easy:
	push ds
	push dx
	lds dx, [ si + 2 ]		; get what we stored in the entry
	mov ah, 25h			; easy case - just reset to the value stored
	int 21h				; doesn't alter CF (leaves NC from unhookinterruptsim) or sets NC
	pop dx
	pop ds
.ret:
	retn

		; INP:	ds:si-> IISP interrupt entry
		;	al = interrupt number
		; OUT:	NC if no error (either hard or easy case),
		;	 ZR if easy case,
		;	  ds:si-> our IISP interrupt entry, containing stored interrupt
		;	 NZ if hard case,
		;	  ds:si-> our IISP interrupt entry
		;	  es:di-> IISP interrupt entry to modify
		;	CY if error (not first handler and no IISP chain to this handler)
		; CHG:	ah, es
rxansi_unhookinterruptsim:
	push bx
	mov ah, 35h			; get current vector
	int 21h				; es:bx-> current interrupt handler
	cmp si, bx			; our pointer ?
	jne .hard

	push ax
	push si
	mov si, ds
	mov ax, es
	 cmp si, ax			; our segment ?
	pop si
	pop ax
	jne .hard

	and ah, 00h			; NC, ZR
	pop bx
	retn
.hard:
		; INP:	ds:si-> IISP interrupt entry
		;	es:bx-> current interrupt entry
		; OUT:	CY if error
		;	NC, NZ if no error,
		;	 ds:si-> our IISP interrupt entry
		;	 es:di-> IISP interrupt entry to modify
		; CHG:	ah, es
	cmp word [es:bx +6], "KB"	; (signature) "KB"/424Bh ? ("BK" in MASM)
	jne .fail
	cmp word [es:bx], 0EA90h	; (jump to next entry) nop\jmp far imm16:imm16 ?
	je .pass			; unused IISP entry (created by iHPFS) -->
	cmp word [es:bx], 10EBh		; (jump to entry's actual code) jmp short $+12h ?
	jne .fail
					; [ es:bx + 2 ] = 4 byte next entry
	cmp byte [es:bx +9], 0EBh	; (jump to far/int call "hardware reset") jmp short ... ?
					; [ es:bx + 11 ] = 7 byte "reserved" (don't look at it)
	jne .fail

.pass:
	cmp [es:bx +2], si		; our pointer ?
	jne .next			; no, try next -->
	push si
	mov si, ds
	 cmp [es:bx +4], si		; our segment ?
	pop si
	je .found			; yes -->
.next:
	les bx, [es:bx +2]		; get next interrupt entry
	jmp short .hard

.found:				; found reference to our interrupt handler
	mov di, bx			; es:di-> IISP interrupt entry that references our's
	or ah, 0FFh			; NC, NZ
	pop bx
	retn

.fail:				; IISP incompatible TSR between current interrupt entry and our entry
	stc
	pop bx
	retn
%endif

%ifn _RXANSI_NOI2F
				; i2F to prevent other ANSI(.SYS) from installing
iispentry rxansi_i2f, rxansi_i2d
	cmp ah, 1Ah			; 1Ah is ANSI.SYS services
	jne .chain
	test byte [cs:rxansi_ctrl0], 80h	; ANSI active ?
	jz .chain			; no -->
	cmp al, 00h			; 0 is installation check
	jz .installationcheck
	cmp al, 02h			; both internal ANSI.SYS functions
	jbe .returnerror		;  we'll just return an unknown error -->
					; Like RBIL says about "DOS" 5+ ANSI.SYS, chain on al>2
					; (Some MS-DOS ANSI.SYS are buggy and compare signed
					;  so they'll return with al=0 or so on al>=80h)
.chain:
	jmp far [cs:.next]

.installationcheck:
	mov al, 0FFh
	iret

.returnerror:				; Set interrupt caller's flags to CY, return with ax=0
	push bp
	mov bp, sp
			; [bp +0]	= bp
			; [bp +2]	= ip
			; [bp +4]	= cs
			; [bp +6]	= fl
	or byte [bp +6], 01h		; CY
	pop bp
	xor ax, ax
	iret
%endif

%if _RXBIO || _RXANSI_RESUNINST
iisphwreset rxansi_i2d
%endif
iispentry rxansi_i2d, rxansi_i2d

	amisinsure

	cmp ah, __AMIS_NUMBER__
rxansi_amisnum equ $-1			; AMIS multiplex number (data for cmp opcode)

	amisincrease

	je .handle			; our multiplex number -->
	jmp far [cs:.next]		; else go to next handler -->

.handle:
	cmp al, 00h
	je .installationcheck		; installation check -->
	cmp al, 02h
	je .uninstall			; uninstallation -->
	cmp al, 04h
	je .determineinterrupts		; determine hooked interrupts -->
	cmp al, 10h
	je .getver			; get compatible version -->
	cmp al, 20h
	je .getctrl0			; get ctrl0 -->
				; all other functions are reserved or not supported by RxANSI
.nop:
	mov al, 0			; show not implemented
	iret

.installationcheck:
	mov al, 0FFh			; Show we're here
	mov cx, RXANSI_VERP		; cx = version
%if _RXBIO
	mov di, rxansi_amissig		; dx:di -> AMIS signature strings of this program
%else
	xor di, di
%endif
.iret_dx_cs:
	mov dx, cs

.iret:
	iret

.uninstall:
	d4bp
%if _RXBIO
		; The internal version can't uninstall, so we'll always report
		; that the uninstallation was unsuccessful. (You could call that
		; lying, but it's AMIS compliant. Reporting not implemented
		; would be AMIS incompliant but "honest".)
	dec ax ; mov al, 01h		; (al is 02h on entry) unsuccessful
	iret
%elif _RXANSI_RESUNINST
		; Resident uninstallation code (only if not the internal version
		; and requested). However this requires some bytes of memory, and
		; if one or more of the interrupts fail to unhook, the code can't
		; display which interrupt(s) failed. (However, this made the unhook
		; loop code shorter.)
		;
		; The transient program (RXANSIT) will always include it's own
		; uninstallation code so it can uninstall any installed RxANSI,
		; even if that wasn't installed by the same version.
	push ds
	push es
	push di
	push si
	push ax
	push dx
	push bx				; AMIS doesn't explicitly list any registers are allowed to
					; change. So just preserve all (except al, gets return code).
	cld
	push cs
	pop ds
	mov bx, rxansi_amisintr		; ds:bx-> interrupt list
	push bx
	mov dx, rxansi_unhookinterruptsim
	call rxansi_loopamisintr
	pop bx				; restore ds:bx->
	jc .uninstall_done

	mov dx, rxansi_unhookinterrupt
	call rxansi_loopamisintr

.uninstall_done:			; CF set according to state!
	pop bx
	pop dx
	pop ax
	pop si
	pop di
	pop es
	pop ds

	dec ax ; mov al, 01h		; (al is 02h on entry; CF left unchanged) suppose error
	jc .iret
	mov al, 0FFh			; uninstalling successful
		; The task is now to free the memory we're executing in. All
		; registers are set for returning to the interrupt caller.
		;
		; This code demonstrates a very special method to insure it
		; won't execute any code in unallocated memory. That's actually
		; the only reason I wrote it. Great if someone could show me an
		; even fancier way of doing this ;-)
%if 0
State is now like when entering any interrupt:

[sp+0] = ip
[sp+2] = cs
[sp+4] = fl

%endif
	push bp
%if _386
	push dword (8+2+6)<<16| 0CA07h
	push dword 589Dh<<16| 21CDh
%elif _186
	push word 8+2+6
	push word 0CA07h
	push word 589Dh
	push word 21CDh
%else ;_186
	mov bp, 8+2+6			; 8 = code; 2 = bp;  6 = fl, cs, ip
	push bp
	mov bp, 0CA07h			; pop es\retf {imm16} - imm16 is the former pushed value
	push bp
	mov bp, 589Dh			; popf\pop ax
	push bp
	mov bp, 21CDh			; int 21h
	push bp
%endif ;_186
		; the actual code is:
		;	int 21h
		;	popf
		;	pop ax
		;	pop es
		;	retf 8+4+6
	mov bp, sp			; ss:bp-> code
%if _386
	push dword [bp+8+2 +0]
%else ;_386
	push word [bp+8+2 +2]		; cs
	push word [bp+8+2 +0]		; ip
%endif ;_386
	push es				; es
	push ax				; ax
	push word [bp+8+2 +4]		; fl
	push ss				; new cs
	push bp				; new ip
	xchg bp, word [bp+8+0]		; get stacked bp
	push cs
	pop es
	mov ah, 49h			; free memory
rxansi_i29.hwreset equ $	; (retf opcode re-used as hardware reset)
	retf				; execute code on stack
%else
	inc ax ; mov al, 03h		; (al is 02h on entry) Safe to remove but no resident uninstaller
	mov bx, cs			; bx=segment
.hwreset equ $-1		; (second byte of mov bx, cs is same as the retf opcode)
	iret
%endif

.determineinterrupts:
	; mov al, 04h			; (al is already 04h on entry) always returns list
	mov bx, rxansi_amisintr		; dx:bx -> Hooked interrupts list
	jmp short .iret_dx_cs

.getver:
	mov al, 0FFh			; show call supported
	mov cx, RXANSI_VERC		; cx = compatible version
	iret

.getctrl0:
	mov bx, rxansi_ctrl0		; dx:bx -> ctrl0
%if _RXANSI_NOLESS
	and byte [cs:bx], ~40h		; switch off LESS bit
%endif
.getctrl_common:
	mov al, 01h			; size of ctrl0: 1 byte
	jmp short .iret_dx_cs

;.getctrl1:
;	mov bx, rxansi_ctrl1
;	jmp short .getctrl_common	; also 1 byte sized

%if _RXBIO
rxansi_i29:
%elif _RXANSI_RESUNINST
iispentry rxansi_i29, rxansi_i29
%else
iispentry rxansi_i29, rxansi_i2d
%endif
	sti
	push ax
	push bx
	push cx
	push dx
	push ds
	push es
	 push cs
	 pop ds
	mov dx, 40h
	mov es, dx			; BIOS data
	mov dx, [rxansi_last]
	mov bl, 7			; Default color
	test byte [rxansi_ctrl0], 80h	; ANSI active?
	jz .noansi			;  no -->

				; ANSI parser
	 mov bl, [rxansi_attrib]
	 cmp dx, 27*256+'['		; Escape,'[' ? (parsing an ANSI sequence ?)
	 jne .noansiseq			; no -->
	 call rxansi_ansiseq
	 jmp short .exit		; ANSI sequence doesn't need to display anything
.noansiseq:
	 cmp al, 27			; Escape?
	 je .filtout
	 cmp dl, 27			; Former character was escape?
	 jne .ansichr			;  No, display normal character -->
	 cmp al, '['
	 je .filtout			; will start parsing the sequence on the next character
	 push ax
	  mov al, dl
	  call rxansi_putcolour
	  mov ah, 0Eh			; TTY output, display the former 27 (escape) that didn't mark a sequence
	  int 10h
	 pop ax

.ansichr:
%ifn _RXANSI_NOLESS
	test byte [rxansi_ctrl0], 40h	; Parse LESS additionally?
	jz .noless			;  No, display with ANSI color -->

.noansi:
	test byte [rxansi_ctrl0], 40h	; LESS active?
	jz .putout			;  No, display with default TTY output -->

				; LESS parser
.both:	cmp dl, 8			; Backspace?
	jne .noless			;  No overstruck sequence -->
	cmp dh, '_'			; Underscore - Backstep - Character? (underlined)
	jz .underline			;  Display yellow -->
	cmp dh, al			; Character - Backstep - Same character? (bold)
	jnz .noless			;  If not, default (ANSI) color -->

.bold:				; "Bold" is displayed green
	and bl, 0F0h			; Zero foreground color
	or bl, 2			; Add foreground color green
rxansi_colorbold equ ($-1)		; Color for bold (data for or opcode)
	jmp short .noless

.underline:			; "Underline" is displayed yellow
	and bl, 0F0h			; Zero foreground color
	or bl, 14			; Add foreground color yellow
rxansi_colorunderline equ ($-1)		; Color for underline (data for or opcode)

.noless:			; No LESS sequence or fall-through
%else	; _RXANSI_NOLESS
.noansi equ .putout
%endif	; _RXANSI_NOLESS
	call rxansi_putcolour
	cmp al, 10
	jnz .putout
	push	dx
	 call	rxansi_getcursor	;PA: DX
	 cmp	dh, [es:84h]		; last row?
	 jnz	.putout_1		; no, continue normal
	  xor	cx, cx
	  mov	dl, [es:4Ah]		; last column
	  dec	dl
	  mov	bh, [rxansi_attrib]
	  mov ax, 0601h
	  int 10h			; Scroll one line
	pop	dx
	mov	al, 10
	jmp short .filtout

.putout_1:
	pop	dx

.putout:			; Just display with default TTY output if
				;  both ANSI and LESS are inactive
	mov ah, 0Eh
	int 10h				; TTY output

.filtout:			; Don't display but still store into buffer
	 xchg	dh, dl
	 mov	dl, al
	 mov	[rxansi_last], dx
.exit:
%if 0
.delaypatch:	jmp short .nodelay	; patched to test al, imm8
	mov ax, word [ es:6Ch ]	; timer ticks since midnight
	inc ax
	cmp byte [ es:6Ch +2 ], 17h	; almost midnight ?
	jb .delay		; no -->
	cmp ax, 0FE80h		; is count behind midnight value ?
	jb .delay		; no, use value -->
	xor ax, ax		; else use zero
.delay:
	cmp ax, word [ es:6Ch ]	; tick reached ?
	jbe .nodelay		; yes -->
	push ax
	mov ax, 1680h
	int 2Fh			; release multitasker timeslice
	or al, al		; supported ?
	pop ax
	jz .delay		; yes -->
	sti
	hlt			; else wait for next IRQ
	jmp short .delay
.nodelay:
%endif
	pop es
	pop ds
	pop dx
	pop cx
	pop bx
	pop ax
	iret

rxansi_putcolour:	; Display space with colour bl, al = actual character
	cmp al, 7
	jz .return
	cmp al, 8
	jz .return
	cmp al, 13
	jz .return
	cmp al, 10
	jz .return			; these don't need a colour
	push cx
	push ax
	 mov bh,[es:62h]		; Active video page
	 mov cx,1
	 mov ax,0920h			; write character and attribute at cursor position,
	 				;  character = blank
	 int 10h
	pop ax
	pop cx
.return:retn

rxansi_ansiseq:		; Process ANSI sequence
	push si
%if _386
	movzx bx, byte [rxansi_argn]
%else
	 mov bl, [rxansi_argn]
	 xor bh, bh			; Fill bx with byte [rxansi_argn]
%endif
	 add bx, rxansi_argv - 1
	 sub al, '0'			; like cmp al,'0'
	 jb .noarg			;  if below -->
	 cmp al, 10			; Digit?
	 jnb .noarg			;  No -->
	 test byte [rxansi_ctrl0], 20h	; Last character was digit?
	 jnz .second_digit		;  Yes, create number from both digits -->
	 inc byte [rxansi_argn]		;  No, create new argument with one digit
	 inc bx				; point to storage for new argument
	 jmp short .first_digit		;  And store digit -->

.second_digit:
	 mov ah, [bx]			; Get old digit
	 add ah, ah
	 add ah, ah
	 add ah, [bx]
	 add ah, ah			; Multiply old digit by ten: ((old*2*2)+old)*2
	 add al, ah			; add old digit to new digit's value
.first_digit:
	 mov [bx], al			; store number
	 or byte [rxansi_ctrl0], 20h
	 jmp short .ret
.noarg:
	 add al,'0'			; Restore previous value
	 call rxansi_subfunc		; Parse letter if any
.ret:	pop si
	retn

		; OUT:	dx = cursor position of active video page
rxansi_getcursor:
	push bx
%if _386
	movzx bx, byte [es:62h]	; Active video page
%else
	 mov bl,[es:62h]		; Active video page
	 xor bh, bh
%endif
	 add bx, bx
	 mov dx,[es:50h +bx]		; Cursor position inside BIOS data
	pop bx
	retn

rxansi_subfunc:
	mov si, rxansi_functiontable
	call rxansi_dispatcher
	jc .end				; Unknown function
	call rxansi_getcursor
	cmp byte [rxansi_argn], 1	; CY if argn zero
	call word [si]
.end:	xor ax,ax
	mov [rxansi_last], ax		; Remove saved ANSI arguments
	mov [rxansi_argn], ax		; Reset argument counter
	mov [rxansi_argv + 1], ax	; Insure arguments are zero
	and byte [rxansi_ctrl0], ~20h	; Mark end of string
	retn

rxansi_functiontable:
		db ';'			; Ignore semicolon
		dw rxansi_afsemi
		db '='			; Ignore equal sign
		dw rxansi_afequal
		db 'H'			; Set cursor position
		dw rxansi_afh
		db 'f'			; Same as 'H'
		dw rxansi_afh
		db 'A'			; Move cursor up
		dw rxansi_afa
		db 'B'			; Move cursor down
		dw rxansi_afb
		db 'C'			; Move cursor forward (right)
		dw rxansi_afc
		db 'D'			; Move cursor backward (left)
		dw rxansi_afd
		db 's'			; Save cursor position
		dw rxansi_afs
		db 'u'			; Restore cursor position
		dw rxansi_afu
		db 'J'			; Clear screen, reset cursor position
		dw rxansi_afj
		db 'K'			; Clear line from cursor position
		dw rxansi_afk
		db 'm'			; Set video colors
		dw rxansi_afm
		db 'h'			; Set video mode
		dw rxansi_afx
		db 'l'			; Reset video mode
		dw rxansi_afx
		db -1

		; Helper function that adjusts stack
rxansi_afsemi:
	and byte [rxansi_ctrl0], ~20h	; end of string
rxansi_afequal:
	pop si				; adjust stack
	retn				; level above

		; Cursor functions
rxansi_afa:
	sbb	dh, [rxansi_argv]
	jnc	rxansi_setcur
	xor	dh, dh
	jmp short rxansi_setcur

rxansi_afb:
	adc	dh, [rxansi_argv]
	jc	.overflow
	cmp	dh, [es:84h]		; higher than last row ?
	jc	rxansi_setcur		; no -->
.overflow:
	mov	dh, [es:84h]		; overflow, force to last row
	jmp short rxansi_setcur

rxansi_afc:
	adc	dl, [rxansi_argv]
	jc	.overflow
	cmp	dl, [es:4Ah]		; higher than last column ?
	jc	rxansi_setcur		; no -->
.overflow:
	mov	dl, [es:4Ah]
	dec	dl			; overflow, force to last column
	jmp short rxansi_setcur

rxansi_afd:
	sbb	dl, [rxansi_argv]
	jnc	rxansi_setcur
	xor	dl, dl			; underflow, force to zero
	jmp short rxansi_setcur

rxansi_afh:
	mov	dx, word [rxansi_argv]
	xchg	dh, dl			; Need to exchange
	cmp	dh, [es:84h]
	ja	.ret
	cmp	dl, [es:4Ah]
	jb	rxansi_setcur		; valid, set cursor position -->
.ret:
	retn

		; Clear row or screen
rxansi_afj:
				; Here was a check if byte [rxansi_argv] is 2. Removed
				; because ANSI.SYS in fact doesn't check, and it's shorter.
	xor	cx, cx
	mov	dl, [es:4Ah]		; number of rows
	dec	dl
	mov	dh, [es:84h]		; number of columns
	mov	bh, [rxansi_attrib]
	mov	ax, 0600h
	int	10h			; Clear screen
	xor	dx, dx			; Reset cursor to 0/0

rxansi_setcur:
	mov	bh, [es:62h]
	mov ah, 02h
	int 10h				; Set cursor
	retn

rxansi_afk:
	mov	cx, dx
	mov	dl, [es:4Ah]		; number of columns
	dec	dl
	mov	bh, [rxansi_attrib]
	mov ax, 0600h
	int 10h				; Clear rest of line
	retn


		; Store and restore cursor position
rxansi_afs:
	mov	[rxansi_curv], dx	; store cursor position
	retn

rxansi_afu:
	mov	dx, [rxansi_curv]	; restore
	jmp short rxansi_setcur


		; Select colour and attributes
rxansi_afm:
	mov	bx, rxansi_argv		; preset to argv
	mov	cl, [rxansi_argn]
	xor	ch, ch
	jcxz	.ret
.loop:
	 mov	al, [bx]
	 cmp	al, 30
	 jc	.nofgcolour
	 cmp	al, 38
	 jnc	.nofgcolour
	 sub	al, 30
	 push	bx
	  mov	bx, rxansi_BRevTab
	  xlat
	 pop	bx
	 mov	ah, ~ 7
	 jmp short .mani			; New foreground colour
.nofgcolour:
	 cmp	al,40
	 jc	.nobgcolour
	 cmp	al,48
	 jnc	.nobgcolour
	 sub	al,40
	 push	bx
	  mov	bx, rxansi_BRevTab
	  xlat
	 pop	bx
%if _186
	 rol	al, 4
%else
	 push	cx
	  mov	cl, 4
	  rol	al, cl			; Exchange high and low nibble
	 pop	cx
%endif
	 mov	ah, ~ 70h
.mani:
	 call	rxansi_atMani
	 jmp short .exit
.nobgcolour:
	 mov	si, rxansi_subfunctiontable
	 call	rxansi_dispatcher
	 jc	.exit
	 call	word [si]
.exit:
	inc	bx			; next argument
	loop	.loop
.ret:
	retn

rxansi_BRevTab: 	; table to get colour number
	db	0,4,2,6,1,5,3,7

		; sub functions
rxansi_atReset:
	mov byte [rxansi_attrib], 7	; Default colour, no blink
	retn

rxansi_atIntens:
	or byte [rxansi_attrib], 8	; set intensity
	retn

rxansi_atUnderl:
	mov	ax,0F801h		; blue for underline
	jmp short rxansi_atMani		; change attribute

rxansi_atFlash:
	or	byte [rxansi_attrib], 80h
	retn

rxansi_atInvert:
	mov	ax,8870h

rxansi_atMani:
	and	ah, [rxansi_attrib]	; mask off old background colour
	or	al, ah
	mov	[rxansi_attrib], al	; new background colour (foreground colour)
	retn

rxansi_atHide:
	mov	ax,8800h
	jmp short rxansi_atMani	; new foreground colour = background colour

rxansi_subfunctiontable:
	db 0
	dw rxansi_atReset
	db 1
	dw rxansi_atIntens
	db 4
	dw rxansi_atUnderl
	db 5
	dw rxansi_atFlash
	db 7
	dw rxansi_atInvert
	db 8
	dw rxansi_atHide
	db -1

		; Select video mode
rxansi_afx:
	jc .ret				; if argn=0
%if _386
	movzx ax, byte [rxansi_argv]
%else
	mov al, [rxansi_argv]
	mov ah,0
%endif
	int 10h
.ret:	retn


rxansi_dispatcher.loop:
	add si, byte 3			; next table entry
rxansi_dispatcher:		; dispatcher
	cmp	[si], al
	jz	.found
	cmp	byte [si], 0FFh		; End of table? (CY if below)
	cmc				; CY if at end of table (not below)
	jnc	.loop			; else NC (value is below) -->
.found:	inc	si			; wo[si] == function pointer
	retn

	endarea rxansi_handler
