
%if 0

RxANSI: Combined ANSI and LESS driver. AMIS v3.6 compliant.
Public Domain by Henrik Haftmann and Christian Masloch

RxANSIT.ASM - Transient code of RxANSI; command line parser and setup

Read RxANSI.DOC for more information.

Assemble: NASM.EXE -fbin -oRXANSI.COM RXANSIT.ASM

Append "-d_NOI2F" to assemble without interrupt 2Fh hook.
Append "-d_NOLESS" to assemble without LESS parser.
This options also work with the supplied make.bat script.

(The following is a German notice)
Assemblieren: NASM.EXE -fbin -oRXANSI.COM RXANSIT.ASM

Mit "-d_NOI2F" ohne Interrupt-2Fh-Umleitung assemblieren.
Mit "-d_NOLESS" ohne LESS-Funktionen assemblieren.
Mit "-d_LANG='DEU'" deutsche Texte assemblieren.
 COMMAND.COM braucht vielleicht "-d_LANG='DEU'" mit allen Anfuehrungszeichen.
Diese Optionen gelten auch mit dem bereitgestellten make.bat Skript.

Hinweis:	Alle weiteren Kommentare in dieser Datei sind in Englisch.
		Die RxANSI-Dokumentation ist bisher ebenfalls Englisch, bei
		Bedarf kann ich aber auch eine deutsche Version erstellen.
		Kontaktdaten siehe am Ende der Datei RxANSI.DOC.

%endif

	%assign _186 0
	%assign _386 0

%include "CMMACROS.MAC"
%include "AMIS.MAC"

	verdef RXANSI,C, 1,10		; Version number of last compatible
	verdef RXANSI,P, 1,10		; Private version

	numdef NOI2F
	numdef RESUNINST
	numdef NOLESS
	incdef NOI2F,RXANSI_NOI2F
	incdef RESUNINST,RXANSI_RESUNINST
	incdef NOLESS,RXANSI_NOLESS

	numdef UPXPAD			; UPX refuses to compress files <3062 bytes. Pad .COM to that size
	strdef LANG, "ENG"
	numdef RXANSI, 1, 1
	numdef RXBIO
	numdef RXANSI_RESUNINST		; include resident uninstaller
	numdef RXANSI_NOLESS		; don't parse LESS sequences
	numdef RXANSI_NOI2F		; Don't hook interrupt 2Fh

%if _RXBIO
 %error "Include RXANSI.ASM for the internal version"
%endif


		; Bits of the option selector (hold in bp)
_holdenvironment	equ 0001h
_forcetoprelocation	equ 0002h
_installlow		equ 0004h
_relocatepsp		equ 0008h
_installnew		equ 0010h
_uninstall		equ 0020h
_lessstate		equ 0040h		; must match LESS bit in ctrl0
_ansistate		equ 0080h		; must match ANSI bit in ctrl0
_state			equ _ansistate|_lessstate
_lessset		equ _lessstate<<2
_ansiset		equ _ansistate<<2
_foundresident		equ 1000h
_incompatresident	equ 2000h
_psprelocated		equ 4000h
_mcbcorruptedcritical	equ 8000h

	[map all RxANSI.MAP]

	cpu 8086
	org 100h

	section TRANSIENT align=1	; Transient program part (command line parser, setup)

transient:
	cld
	mov dx, msg.name
	call disp_msg		; Start with message

			; The UMB link and memory allocation strategy is now always safed here
			; and restored by abort_msg as well as any other termination code.
			; Otherwise, if code below relocatepsp or install.relocate jumped to
			; abort_msg the link/strategy won't be restored.
	mov ax, 5802h
	int 21h
	xor ah, ah		; some "DOS" only return al
	mov word [ cs:restored.umblink ], ax	; save UMB link

	mov ax, 5800h
	int 21h
	mov word [ cs:restored.strategy ], ax	; save strategy

	mov dx, i23
	mov ax, 2523h
	int 21h			; set Int23 so that we restore link/strategy on Ctrl-C as well
				; (We could also include some sort of Int24 so that critical error
				;  responses of "Abort" would also restore link/strategy. However I
				;  think that's really not required because RxANSI doesn't really
				;  make any DOS call that accesses critical error throwing devices.)

	xor bp, bp		; initial option selector

	mov ax, cs
	add ax, 10h+transient_size_p
	mov ds, ax
	mov bx, rxansi_ctrl0	; ds:bx -> resident segment's ctrl0

findinstalled:
	mov ah, 0FFh		; start with multiplex number 0FFh
.loop:
	mov al, 00h		; AMIS installation check
	int 2Dh			; AMIS (or "DOS reserved" = iret if no AMIS present)
	cmp al, 0FFh
	je .try
.notfound:
	sub ah, 1		; search is backward (to find latest installed first), from 0FFh to 00h including
	jnc .loop		; try next if we didn't checked all yet -->

	jmp short scancl	;  If not found, start command line processing -->

.try:
	mov si, rxansi_amissig	; ds:si -> our AMIS name strings
	mov es, dx		; es:di -> name strings of AMIS multiplexer that just answered
	mov cx, 16		; Ignore description, only compare vendor and program name
	repe cmpsb
	jne .notfound		; No match, try next
	mov ds, dx		; set ds to our resident copy's segment
	mov al, 10h		; get compatible RxANSI version
	int 2Dh
	cmp al, 0FFh		; al = 0FFh if supported
	jne .unknown		; doesn't support this question -->
	cmp cx, RXANSI_VERC	; Compatible version?
	jne .unknown
	mov al, 20h		; Get RxANSI ctrl0
	int 2Dh
	cmp al, 00h		; al = size of returned ctrl data. at least one byte if supported
	jne .known		; if size ok, known -->
.unknown:	
	or bp, _incompatresident; indicate resident copy is incompatible
.known:
	or bp, _foundresident	; indicate resident copy found

scancl:			; bp = selected options
			; ah = AMIS multiplex number of resident copy if bp&_foundresident (else 0)
			; ds:bx-> ctrl0 of either resident copy or stored handler

	mov si, 81h		; PSP command line
	push cs
	pop es			; es changed while searching for resident
.loop:	cs lodsb		; ds is resident segment, so use cs
	cmp al, 9		; tab ?
	je .loop
	cmp al, 32		; blank ?
	je .loop
	jb .end			; zero, cr, lf ? -->
	cmp al, '-'
	je .switch
	cmp al, '/'
	jne .invalid		; unknown character -->
			; parse switch
.switch:
	cs lodsb		; get next character
	cmp al, 'a'
	jb .notlowercase
	cmp al, 'z'
	ja .notlowercase
	and al, ~ 20h		; Uppercase if required
.notlowercase:
			; check against known switches
	push si
	push ax
	mov si, parameters
	xchg al, ah
.loopswitch:
	cs lodsb
	xchg ax, cx
	cs lodsb
	xchg ax, cx		; cx = option selector bits
	cmp al, 0FFh
	je .invalid		; not found!
	cmp al, ah
	jne .loopswitch		; try next -->
	pop ax			; %$&" preserve ah (multiplex number) !
	pop si
	xor ch, ch

	test cl, _state		; ANSI/LESS state ?
	jz .normalbp		; no, normal option --> (sets bit in bp)

		; cx = bit to set or clear in bp
		; cx<<2 = bit to set in bp to show the cx bit was set
		; ds:si-> '+' to set, '-' to clear
	cs lodsb
	cmp	al,'-'		; OFF ?
	jne	.notminus
	call	flagset
	not	cx		; reverse bits
	and	bp,cx		; do bp&!cx
	jmp short .loop
.notminus:
	cmp	al,'+'		; ON ?
	jne	.invalid
	call	flagset
				; then fall through to do bp|cx
.normalbp:
	or bp, cx
	jmp short .loop		; ds:si-> character after switch character. just continue

.invalid:			; display help if invalid
	mov dx, msg.help
	jmp short .abort_msg

.end:
	test bp, _uninstall
	jz .dontuninstall
	test bp, _forcetoprelocation|_holdenvironment| \
		_installlow|_installnew|_relocatepsp| \
		_ansiset|_lessset
	jnz .invalid		; all invalid if /u selected
	jmp uninstall

.dontuninstall:
	test bp, _foundresident	; no resident found ?
	jz .install		; yes, install -->
	test bp, _installnew	; forced to install new ?
	jz .dontinstall		; no -->

.install:
	jmp install

.dontinstall:
	test bp, _forcetoprelocation|_holdenvironment| \
		_installlow|_relocatepsp
	jz .changeflags
	mov dx, msg.already	; can't install new, already resident
	jmp short .abort_msg

.changeflags:
	test bp, _incompatresident
	jz .compatible
	mov dx, msg.incompatible ; can't change flags of incompatible version
.abort_msg:
	jmp abort_msg

.compatible:
	push ax
	call changeflags
	pop ax			; AMIS multiplex number
	mov al, 20h
	int 2Dh			; clear unsupported bits

		; Display status
	mov dx, msg.stateansi	; "state ANSI: "
	call disp_msg
	test byte [bx],80h
	call onoff		; "ON (+)" or "OFF (-)" depending on ZF
	mov dx, msg.stateless	; ", LESS: "
	call disp_msg
	test byte [bx],40h
	call onoff
	mov dx, msg.crlf
	call disp_msg
	call restorestate
	mov ax, 4C00h
	int 21h

flagset:
	push cx
	rol cx, 1
	rol cx, 1
	or bp, cx
	pop cx
	retn


		; INP:	al = interrupt number
		;	ds:si-> interrupt entry
		; OUT:	CY if unhooking failed
		;	NC if unhooking successful
		; CHG:	ah, es, di
unhookinterrupt:
			; unhookinterruptsim (below) only checks if it's possible to unhook this interrupt.
			; This function really unhooks the interrupt if possible.
			;
			; This is to cover the situation when some of the hooked interrupts can unhook,
			; but some can't. If the uninstaller would start to unhook the interrupts and then
			; catch the interrupt that can't be unhooked the user would end with a dead TSR
			; that's uninstalled halfway. Very bad.
			;
			; "Simulating" the unhooking first and checking if all Interrupts can unhook
			; normally will not return such a state.
	call unhookinterruptsim
	jc .ret				; bad. -->
	jz .easy
.hard:
				; "hard" case: unhookinterruptsim has however already done the work,
				; so the hard case is here indeed easier than the easy case.
	cmpsw				; skip 10EBh or 0EA90h
	movsw				; copies our stored pointer into the other's entry
	movsw				; copies our stored segment into the other's entry
	clc
	retn
.easy:
	push ds
	push dx
	lds dx, [ si + 2 ]		; get what we stored in the entry
	mov ah, 25h			; easy case - just reset to the value stored
	int 21h				; doesn't alter CF (leaves NC from unhookinterruptsim) or sets NC
	pop dx
	pop ds
.ret:
	retn

		; INP:	ds:si-> IISP interrupt entry
		;	al = interrupt number
		; OUT:	NC if no error (either hard or easy case),
		;	 ZR if easy case,
		;	  ds:si-> our IISP interrupt entry, containing stored interrupt
		;	 NZ if hard case,
		;	  ds:si-> our IISP interrupt entry
		;	  es:di-> IISP interrupt entry to modify
		;	CY if error (not first handler and no IISP chain to this handler)
		; CHG:	ah, es
unhookinterruptsim:
	push bx
	mov ah, 35h			; get current vector
	int 21h				; es:bx-> current interrupt handler
	cmp si, bx			; our pointer ?
	jne .hard

	push ax
	push si
	mov si, ds
	mov ax, es
	 cmp si, ax			; our segment ?
	pop si
	pop ax
	jne .hard

	and ah, 00h			; NC, ZR
	pop bx
	retn
.hard:
		; INP:	ds:si-> IISP interrupt entry
		;	es:bx-> current interrupt entry
		; OUT:	CY if error
		;	NC, NZ if no error,
		;	 ds:si-> our IISP interrupt entry
		;	 es:di-> IISP interrupt entry to modify
		; CHG:	ah, es
	cmp word [es:bx +6], "KB"	; (signature) "KB"/424Bh ? ("BK" in MASM)
	jne .fail
	cmp word [es:bx], 0EA90h	; (jump to next entry) nop\jmp far imm16:imm16 ?
	je .pass			; unused IISP entry (created by iHPFS) -->
	cmp word [es:bx], 10EBh		; (jump to entry's actual code) jmp short $+12h ?
	jne .fail
					; [ es:bx + 2 ] = 4 byte next entry
	cmp byte [es:bx +9], 0EBh	; (jump to far/int call "hardware reset") jmp short ... ?
					; [ es:bx + 11 ] = 7 byte "reserved" (don't look at it)
	jne .fail

.pass:
	cmp [es:bx +2], si		; our pointer ?
	jne .next			; no, try next -->
	push si
	mov si, ds
	 cmp [es:bx +4], si		; our segment ?
	pop si
	je .found			; yes -->
.next:
	les bx, [es:bx +2]		; get next interrupt entry
	jmp short .hard

.found:				; found reference to our interrupt handler
	mov di, bx			; es:di-> IISP interrupt entry that references our's
	or ah, 0FFh			; NC, NZ
	pop bx
	retn

.fail:				; IISP incompatible TSR between current interrupt entry and our entry
	stc
	pop bx
	retn


..@stack:			; above is overwritten by stack while installing
abort_msg:
	call disp_msg
	call restorestate
	mov ax, 4CFFh
	int 21h

changeflags:
		; bp = option selector
		; ds:bx-> ctrl0
	mov ax, bp
	test ah, _ansiset>>8
	jz .noansi
	push ax
	and al, _ansistate
	and byte [bx], ~_ansistate
	or byte [bx], al
	pop ax
.noansi:
	test ah, _lessset>>8
	jz .noless
	and al, _lessstate
	and byte [bx], ~_lessstate
	or byte [bx], al
.noless:
	retn

uninstall:
	d4bp
	test bp, _foundresident
	jnz .found
	mov dx, msg.notinstalled	; nothing to uninstall there
.abort_msg:
	jmp short abort_msg

.found:
	test bp, _incompatresident
	jz .compat
	mov dx, msg.incompatible	; resident, but incompatible version
	jmp short abort_msg

.compat:
			; ds:bx-> ctrl0 of resident copy (unused yet)
			; ah = AMIS multiplex number of resident copy
	mov al, 02h
	mov dx, cs
	mov bx, .done			; dx:bx = return address if successful
	int 2Dh				; AMIS uninstall TSR
	cmp al, 0FFh			; 0FFh successful
	je .done			; TSR has already done anything -->
	cmp al, 00h			; 00h not implemented
	je .continue_noseg		; do it myself -->
	cmp al, 03h			; 03h safe, no resident uninstaller (still enabled). bx=segment
	je .continue_seg		; expected -->
	cmp al, 04h			; 04h safe, no resident uninstaller (now disabled). bx=segment
	je .continue_seg		; unexpected, but continue -->
		; (other values: 01h unsuccessful or internal, 02h can't uninstall yet,
		; but will do so when able, 05h not safe, 06h,07h device driver; all fail)
.fail:
	mov dx, msg.cantuninstall	; some error occured
	call disp_msg
	mov dl, '.'
	mov ah, 02h
	int 21h				; display dot to end message
	mov dx, msg.crlf
	jmp short .abort_msg

.continue_noseg:
	xor bx, bx			; = 0

.continue_seg:
	push bx

	mov al, 04h
	int 2Dh				; AMIS determine hooked interrupts
	cmp al, 04h
	jne .fail
	mov ds, dx			; ds:bx-> interrupt table
	push bx
	mov dx, unhookinterruptsim
	mov cx, .return1
	jmp short loopamisintr

.return1:
	jc .unhookerror
	pop bx				; restore ds:bx->

	mov dx, unhookinterrupt
	mov cx, .return2
	jmp short loopamisintr

.return2:
	jc .unhookerror

	push ds
	pop es				; es = segment of interrupt list
	pop ax				; stacked segment
	or ax, ax			; marker to use interrupt list segment ? (zero)
	jz .uselistsegment		;  yes -->
	mov es, ax			; else use segment returned by uninstall function

.uselistsegment:
	mov ah, 49h			; Free memory
	int 21h
				; Note:	General deinstallation code should not assume the TSR was just in
				;	that single memory block. If the deinstallation code doesn't know
				;	the TSR, it should search (and if it founds any, free) memory blocks
				;	that have the MCB owner either that's now in es (code segment of TSR)
				;	or the same MCB owner value that the TSR's code segments MCB had.
				;	(Only if the TSR's code segment was a valid memory block with MCB.)
	jc .fail			; if that causes an error, still report "fail"
					; (but interrupts are already unhooked now)

.done:
	mov dx, msg.removed
	call disp_msg			; "removed."
..@exit:
	call restorestate
	mov ax, 4C00h
	int 21h

.unhookerror:
	mov dx, msg.cantuninstall
	call disp_msg

	 pop dx
	cmp dx, byte -1			; at least another interrupt failed ? (at least two interrupts failed ?)
	 push dx
	je .unhookerrorsingle
	mov dx, msg.interrupts		; first message for multiple interrupts
	jmp short .unhookerrorfirst
.unhookerrorsingle:
	mov dx, msg.interrupt		; first message for single interrupt
.unhookerrorfirst:
	mov bx, dx

.unhookerrorloop:
	call disp_msg			; msg.interrupts or msg.hcomma
	call disp_al_hex
	pop ax				; get next value from stack
	 pop dx
	cmp dx, byte -1			; at least two more ?
	 push dx
	mov dx, msg.hcomma		; message if at least two more interrupts to display
	jne .unhookerrornotlast		; yes -->
	mov dx, msg.hand		; only one more interrupt to display
.unhookerrornotlast:
	cmp ax, byte -1			; last error marker?
	jne .unhookerrorloop		; no, keep looping -->

	mov dx, msg.hooked		; last message for single interrupt
	cmp bx, msg.interrupt		; was first message for single interrupt ?
	je .unhookerrorabort		; yes -->
	mov dx, msg.hookeds		; else use last message for multiple interrupts
.unhookerrorabort:
	jmp abort_msg

		; INP:	ds:bx -> AMIS interrupt list
		;	dx = code called for each interrupt (with ds:si -> interrupt entry)
		;	cx = code to return to
		; OUT:	CY if any interrupt failed,
		;	 ax = first failed interrupt
		; 	 ss:sp-> 0FFFFh terminated list of failed interrupts on stack
		;	NC if all interrupts successful uninstalled
		; CHG:	si, ax, bx, what called code (dx) changes
		;
		; Called code (dx) has to preserve dx, bx, cx, al.
loopamisintr:
	
	mov ax, 0FFFFh
	push ax
.loop:
	mov al, [bx]
	mov si, [bx +1]

	call dx
	jnc .noerror			; no error -->
	xor ah, ah
	push ax				; else remember number of interrupt, but continue looping

.noerror:
	add bx, byte 3
	cmp al, 2Dh
	jne .loop			; do until 2Dh done -->

	pop ax
	cmp ax, byte -1			; if it's below (CY), there were errors
	jmp cx



install:		; ds = handler segment inside cs
			; bp = option selector
	d4bp
	mov ax, cs
	add ax, 10h+transient_size_p
	mov ds, ax		; ds = handler segment (could have changed if /n switch and already resident)

	mov ah, 0Fh
	int 10h			; get video mode etc, bh = active video page
	mov ah, 08h		; read character and attribute at cursor position
	int 10h
	mov byte [rxansi_attrib], ah	; inherit current attribute

	mov bx, rxansi_ctrl0	; ds:bx -> ctrl0
	call changeflags	; set flags from bp


	test bp, _holdenvironment ; don't free environment ?
	jnz .dontfreeenvironment ; yes -->

	mov ah, 49h
	mov es, word [ cs:2Ch ]
	int 21h			; Free our environment
.dontfreeenvironment:

	mov sp, ..@stack	; change to stack inside now unused code

	mov ah, 4Ah
	push cs
	pop es			; Resize the segment allocated to this program
	mov bx, 10h+transient_size_p+rxansi_handler_size_p	; 10h is the size of the PSP
	int 21h

.allocate:
			; The DOS memory allocation strategy classes are actually "Low then high",
			; "High only" and "High then low". RBIL tells trash regarding that. We've
			; to disable the UMB link to insure only low memory is allocated. Normally,
			; the UMB link is already disabled when the program starts. But some
			; programs, like CMDEDIT 3.21, are buggy and leave the UMB link enabled
			; after they used UMBs.
	mov bx, 1		; enable UMB link
	mov dx, 0081h		; best fit, high then low strategy
	test bp, _installlow	; use settings for high ?
	jz .high		; indeed -->
	mov bl, 0		; disable UMB link (bh already 0)
	mov dl, 01h		; best fit, low then high strategy (dh already 0)

.high:
	mov ax, 5803h
	int 21h			; set UMB link (disable for low, enable for high)

	mov ax, 5801h
	mov bx, dx
	int 21h			; set strategy (low then high for low, high then low for high)
	jnc .highsupported
			; If the current "DOS" is < 5 (or: doesn't support UMBs),
			; the low then high allocation strategy is used instead:

	mov ax, 5801h		; filled with error code, reset to function number
	mov bl, 01h		; set to best fit, low then high strategy
	int 21h			; set strategy
.highsupported:

	mov ah, 48h
	mov bx, rxansi_handler_size_p	; required size (paragraphs)
	int 21h
	jnc .memoryallocated
	mov dx, msg.mcbcorrupted
	cmp ax, 0007h			; MCB chain corrupted error code ?
	je .abort_msg			; yes -->
	mov dx, msg.nomemory		; else assume not enough memory
.abort_msg:
	jmp abort_msg

		; ax = allocated memory block
		; ds = segment of resident (behind transient)
.memoryallocated:
	mov es, ax
	d4bp
		; es = ax
	test bp, _psprelocated
	jnz .dontrelocatepsp
	push ax
	mov ax, 335Eh
	int 21h			; (changes ax,bx,cx,dx,si,di)
	cmp ax, "Rx"		; RxDOS ? (7.30+)
	pop ax
	je .relocatepsp
			; Problem: What if there's a TSR that somehow tracks PSPs (or allocates
			; resources to PSP addresses)?
	test bp, _relocatepsp	; PSP relocation allowed ?
	jz .dontrelocatepsp	; no -->
.relocatepsp:
	call relocatepsp
	jnc .allocate
.dontrelocatepsp:
	test bp, _forcetoprelocation
	jnz .relocate		; don't test if behind program -->
	mov bx, cs
	add bx, 10h+transient_size_p+rxansi_handler_size_p +1
	cmp ax, bx		; did we allocate the block just behind the PSP block ?
	je .relocate		; yes, relocate -->
.dontrelocate_1:		; else no need to relocate
	jmp .dontrelocate

.relocate:

%if 0

If the allocated block is located behind the PSP block, a gap of the PSP size, transient
size and resident size (10h+transient_size_p+rxansi_handler_size_p) would be created by
using that block. Thus the allocated block is freed, the transient code is relocated into
a newly allocated block and the resident code is relocated to where the transient code
was before (behind the PSP). The PSP and the relocated transient code are both freed by
the termination but the newly created MCB of the relocated resident code (at cs+10h)
isn't. This still leaves a gap, but it's just 256 byte sized (PSP size).

This relocation helps to decrease the "gap size" if the program is loaded in the LMA and
then installs the resident code there as well (as usual if there's no UMA). The PSP
itself isn't relocated. (Relocating the PSP wouldn't create any gap at all, but it may
not work on all systems. And you've to change all allocated MCBs and the current DOS PSP
manually.) Those "good old" TSRs don't create any gap. But it's better to have a 256 byte
gap than to hold a useless PSP resident in the same gap.

If, however, the free memory above the allocated memory is _smaller_ than the gap created
by relocating the resident part behind the PSP, it's instead relocated to the end of that
memory so that no gap stays in memory.

Some examples:

Seg.  Usage
0000: 64 KiB used by DOS and other programs
1000: PSP
1010: Transient code
1090: Resident code
10C7: Free memory	<-- Memory allocated first is in this block
The transient code is relocated to somewhere else. The resident code is copied down from
segment 1090h to 1011h (at 1010h, there's the newly created MCB). The memory behind the
resident code is freed; so (with installing Int2F) memory from 1039h is free. This
results in the following memory map after the transient code terminated:
0000: 64 KiB used by DOS and other programs
1000: Gap (size of the PSP)
1010: Resident code
1048: Free memory

Seg.  Usage
0000: 64 KiB used by DOS and other programs
1000: PSP
1010: Transient code
1090: Resident code
10C7: Free memory	<-- Memory allocated first is in this block
1100: Used by another program
1300: Free memory (up to 9FFF)
The resident code is relocated to 10C8h instead, because the gap between the resident
code and the other program above is then filled. This results in the following memory
map:
0000: 64 KiB used by DOS and other programs
1000: Gap
10C8: Resident code
1100: Used by another program
1300: Free memory (up to 9FFF)

%endif

	d4bp
	push ds
	mov ah, 4Ah		; resize memory
	mov bx, 0FFFFh		; will not succeed
	int 21h			; but returns largest possible in bx
			; Note:	DOS can actually change the size to the largest possible!
	mov dx, msg.mcbcorrupted
	jnc .abort_msg		; succeeded?!
	sub bx, byte 10h
	test bp, _forcetoprelocation
	jnz .relocate_beforenext
	cmp bx, rxansi_handler_size_p
	ja .relocate_behindpsp

.relocate_beforenext:
	d4 push dx
	d4 mov dx, msg.relocbeforenext
	d4 call disp_msg
	d4 pop dx
	d4bp
	add bx, byte 10h
	mov ah, 4Ah
	int 21h			; now really resize to largest possible
	sub bx, rxansi_handler_size_p
	mov ax, es
	jmp short .relocate_common

.relocate_behindpsp:
	d4 push dx
	d4 mov dx, msg.relocbehindpsp
	d4 call disp_msg
	d4 pop dx
	d4bp
	mov ah, 49h
	int 21h			; free memory in es (allocation)
	mov ah, 48h
	mov bx, transient_size_p+rxansi_handler_size_p
	int 21h			; allocate memory for relocation of transient and resident
.abort_msg_c:
	jc .abort_msg
	mov es, ax
	xor di, di
	mov si, transient	; = 100h
	mov cx, words(fromparagraphs(transient_size_p+rxansi_handler_size_p))
	repz cs movsw		; copy transient from cs:si-> to es:di-> (including unused code as stack)
	sub ax, byte 10h	; adjust because transient expects a PSP in front of it
	push ax			; new ss
	push ax			; new cs
	mov ax, .relocated
	push ax			; new ip
	mov ax, cs		; pass to work on it after relocation
	retf

.relocated:
	pop ss			; change stack to relocated transient (sp is still valid!)

	pop bx
	mov bx, cs
	add bx, 10h+transient_size_p
	push bx			; new ds to restore

	mov bx, 10h +1

	push ax
	 push cs
	 pop ds
	mov dx, i23
	mov ax, 2523h		; set Int23 to relocated position
	int 21h
	pop ax

		; ax = allocated block
		; bx = paragraph of allocated block used
		;
		; ax: PSP/lower part of allocated block
		; ax+bx: free memory (former transient)/upper part of allocated block
		;
		; This code splits the memory block at ax in two blocks. The lower block
		; (the PSP/unused) is freed by the termination of the transient code.
.relocate_common:
	dec ax
	mov ds, ax		; former transient's MCB
	add ax, bx
	mov es, ax		; just behind the PSP
	xor di, di
	xor si, si
	mov cx, 8
	repz movsw		; copy MCB at ds:si-> to es:di->
	sub word [es:3], bx	; adjust size to next
	jc .abort_msg_c
	mov word [es:5], cx
	mov byte [es:7], cl	; force the unused bytes to zero
	cmp byte [0], 'M'	; 'M' ?
	je .validmcb		; all right -->
	cmp byte [0], 'Z'	; not 'Z' ?
	jne .abort_msg		; error -->

.validmcb:
	dec bx
	mov word [3], bx	; it doesn't hurt if it's still 'Z' here
	mov byte [0], 'M'	; the created MCB always follows this one

	inc ax
	mov es, ax		; memory block of created MCB
	pop ds

	mov bx, rxansi_handler_size_p
	mov ah, 4Ah
	int 21h
	jc .abort_msg_c
		; ax = created memory block (behind PSP)
		; es = segment of memory block that contains resident

.dontrelocate:
			; es = segment of allocated block
			; ds = where to copy resident from
	mov cx, rxansi_handler_size_w
	xor di, di		; es:di-> start of allocated block
	xor si, si		; ds:si-> source inside this block
	repz movsw

	mov ah, 00h		; start with multiplex number 00h
.loopamis:
	mov al, 00h		; installation check
	int 2Dh
	cmp al, 00h		; still 00h if free
	jz .foundamis
	inc ah			; search from bottom to top
	jnz .loopamis		; loop if zero not yet reached -->
	mov dx, msg.noamisnumber; abort if no free AMIS multiplex number left
	jmp abort_msg

.foundamis:
			; ah = free AMIS multiplex number
	mov byte [ es:rxansi_amisnum ], ah	; Set multiplex number

			; All error conditions must be checked above because we'll now
			; modify the MCB so that it won't be deallocated when terminating.
	mov ax, es
	dec ax
	mov ds, ax		; ds = MCB of allocated block
	inc ax			; ax = allocated block!
	mov word [0008h], "RX"
	mov word [000Ah], "AN"
	mov word [000Ch], "SI"
	and word [000Eh],byte 0	; Force MCB string to "RXANSI",0
	mov word [0001h], ax	; Set owner to itself

			; ax = segment of resident copy
			; cx = size of copied handler in words
	cmp ax, 0A000h		; is that high ?
	mov dx, msg.highloadedusing
	jae .highloaded		; of course -->
	mov dx, msg.loadedusing	; no, other message

.highloaded:
	call disp_msg		; "[high ]loaded using "
	push ax			; allocated block
	mov ax, rxansi_handler_size_p*16 +16	; +16 is for the MCB (most MEMs count it)
	call disp_ax_dec	; "<num>"
	mov dx, msg.byte
	call disp_msg		; " bytes."

	pop ds			; ds = allocated block
	mov bx, rxansi_amisintr	; ds:bx -> AMIS interrupt list
	mov dx, hookinterrupt
	mov cx, ..@exit
	jmp loopamisintr	; hook all interrupts in list, return to ..@exit


relocatepsp:
	mov dx, ax
	mov bx, cs
	cmp ax, 0A000h			; allocated in UMA ?
	jae .ispspinuma			; yes -->
.ispspinlma:
	cmp bh, 0A0h			; PSP also in LMA ?
	jb .samearea			; yes -->
.dontrelocate:
	mov ax, dx
	mov es, ax			; insure ax and es set (when branching from below)
	stc				; no, don't relocate
	jmp .return			; sets ds to required value
.ispspinuma:
	cmp bh, 0A0h			; PSP also in UMA ?
	jb .dontrelocate		; no -->
.samearea:
	mov ax, 5803h
	mov bx, 1
	int 21h				; enable UMB link

	push es				; allocated memory block
	mov ah, 52h
	int 21h				; es:bx-> List of Lists
	mov di, [ es:bx -2 ]		; first MCB
	pop dx				; allocated memory block
	mov bx, cs			; PSP and program memory block
				; dx = allocated memory block
				; bx = PSP and program memory block
				; di = first MCB
	xor cx, cx			; size of current block
	xor ax, ax			; flags: 20h allocation found
					;	 10h PSP found
					;	 8 best fit is in block that contains allocation (at start),
					;	 4 best fit is in block that contains PSP,
					;	 2 current block contains allocation,
					;	 1 current block contains PSP
	mov si, -1			; size of best fit
	d4bp
.searchblock:
	mov ds, di			; ds = MCB
	cmp di, 9FFFh			; block in LMA ?
	jb .searchblock_allocatedinlma	; yes -->
.searchblock_allocatedinuma:
	cmp bh, 0A0h			; allocated also in UMA ?
	jb .searchblock_terminate	; no -->
	jmp short .searchblock_samearea
.searchblock_allocatedinlma:
	cmp bh, 0A0h			; (NC if above or equal) allocated also in LMA ?
	cmc				; CY if above or equal
	jc .searchblock_terminate	; no -->
.searchblock_samearea:
	cmp word [ 1 ], byte 0		; free ?
	je .searchblock_freeblock
	inc di
	cmp di, dx
	je .searchblock_allocatedblock
	cmp di, bx
	je .searchblock_pspblock
	dec di
.searchblock_terminate:
	cmp cx, rxansi_handler_size_p +1; available block large enough ?
	jb .searchblock_next_nofree	; no -->
	cmp cx, si			; available block fits better than former best fit ?
	ja .searchblock_notbestfit	; no, not best fit -->
	and al, ~(8|4)			; new best fit - clear former's flags
	test al, 1			; contains PSP ?
	jz .searchblock_fitnonpsp	; no -->
	or al, 4			; remember best fit contains PSP
.searchblock_fitnonpsp:
	test al, 2			; contains allocation ?
	jz .searchblock_fitnonalloc	; no -->
	push di
	sub di, cx			; first MCB of this block
	inc di				; data of first MCB
	cmp di, dx			; address of our allocation ?
	pop di
	jne .searchblock_fitnonalloc	; no, not at beginning -->
	or al, 8			; remember best fit contains allocation at beginning
.searchblock_fitnonalloc:
	mov si, cx			; remember size of best fit
.searchblock_notbestfit:
.searchblock_next_nofree:
	xor cx, cx
	and al, ~(2|1)
.searchblock_next:
	cmp byte [ 0 ], 'M'
	jne .searchblock_end		; if ZR, NC
	add di, word [ 3 ]
	inc di
	jmp short .searchblock

.searchblock_freeblock:
	add cx, word [ 3 ]		; add size of this memory block
	inc cx				; plus its MCB to the current block size
	jmp short .searchblock_next

.searchblock_allocatedblock:
	d4bp
	dec di
	or al, 20h|2			; mark that this is the available block with allocation
	cmp word [ 3 ], byte rxansi_handler_size_p	; size matches what we allocated ?
	jne .mcberror			; no, strange DOS -->
	jmp short .searchblock_freeblock; record/add size of allocation to current block

.searchblock_pspblock:
	d4bp
	dec di
	or al, 10h|1			; current block contains PSP
	cmp word [ 3 ], 10h+transient_size_p+rxansi_handler_size_p	; size matches what we resized it to ?
	jne .mcberror			; no, strange DOS -->
	jmp short .searchblock_freeblock; record/add size of PSP memory to current block


.searchblock_end:
				; si = size of best fit
				; ax = flags
				; bx = PSP
				; dx = allocation
	cmp byte [ 0 ], 'Z'		; does it end with 'Z' ?
	jne .mcberror			; nope, so don't trust anything -->
	xor al, (20h|10h)
	test al, (20h|10h)		; were both set ? (now both clear ?)
	jnz .mcberror			; no, either PSP or allocation not found -->
	cmp si, byte -1			; any fit found ?
	je .mcberror			; no, strange (at all, we found our allocation) -->

	test al, 8			; allocation at start of best fit ?
	jnz .dontrelocate		; yes, no need to relocate -->
	test al, 4			; best fit contains PSP ?
	jz .dontrelocate		; no --> (strange - why didn't DOS allocate the best fit?)
				; Now we know that the best fit would be the block that contains
				; the PSP and that the allocation is currently not at the start
				; of this block. So, yes, we want to relocate the PSP now.

				; We also know that the MCB chain contains 'M' MCBs and terminates
				; with a 'Z' MCB, just as we expect.

	push dx
	d4 mov dx, msg.relocatingpsp
	d4 call disp_msg
	pop es				; former non-optimal allocation
	d4bp
	mov ah, 49h
	int 21h				; free it

	mov ax, 5803h
	xor bx, bx
	int 21h				; disable UMB link
	mov ax, 5801h
	mov bx, 0002h
	int 21h				; set strategy (low then high, last fit)

	mov ah, 48h
	mov bx, 10h+transient_size_p+rxansi_handler_size_p	; size of PSP, transient and resident
	int 21h				; try to allocate 
	jc .cantrelocate		; not possible (?!), restore strategy and link and
					;  allocate resident again (without trying to relocate the PSP)
	mov es, ax			; newly allocated block
	xor di, di
	xor si, si
	mov cx, 80h+transient_size_w+rxansi_handler_size_w
	repz cs movsw			; copy PSP and program into newly allocated block
	mov dx, cs			; dx = old PSP

	push ax				; new cs
	mov bx, .relocated
	push bx				; new ip
	retf
.relocated:
	mov ss, ax

	push ax				; new PSP
	dec ax
	mov ds, ax			; new PSP's MCB
	mov word [0008h], "RX"
	mov word [000Ah], "AN"
	mov word [000Ch], "SI"
	and word [000Eh],byte 0		; force MCB string to "RXANSI",0
					; not really required
	mov ax, 5803h
	mov bx, 1
	int 21h				; enable UMB link

	mov ah, 52h
	int 21h				; es:bx-> List of Lists
	mov ax, [ es:bx -2 ]		; first MCB
	pop bx				; bx = new PSP
.fixupallocated:
	mov ds, ax			; ds = MCB
	cmp word [ 1 ], dx		; belongs to old PSP ?
	jne .fixupnextallocated
	mov word [ 1 ], bx		; yes, link to new PSP
	or bp, _psprelocated		; used here as indicator that we edited a MCB
.fixupnextallocated:
	cmp byte [ 0 ], 'M'		; in-chain signature ?
	jne .nonextallocated		; no -->
	add ax, word [ 3 ]		; size of data
	inc ax				; plus size of MCB, address of next MCB
	jmp short .fixupallocated
.nonextallocated:
	d4bp
	cmp byte [ 0 ], 'Z'		; last signature ?
	je .fixedupallocated		; yes -->
	test bp, _psprelocated		; any MCBs edited ?
	jz .mcberror			; no, just report corrupted MCB chain -->
	mov dx, msg.mcbcritical		; inform user that some fragments may belong to an invalid PSP
	jmp short .mcberrorcritical
.fixedupallocated:
	cmp word [ cs:36h ], dx		; AHT points into (old) PSP ?
	jne .dontfixupaht		; no -->
	mov word [ cs:36h ], bx		; else point it into new PSP
.dontfixupaht:
					; bx = new PSP location
	mov ah, 50h
	int 21h				; set current PSP

					; bx = new PSP location
					; dx = old PSP location
	mov ax, 335Dh
	int 21h				; PSP relocated call-out

	push dx
	 push cs
	 pop ds
	mov dx, i23
	mov ax, 2523h			; set Int23 to relocated position
	int 21h
	pop es				; former PSP
	mov ah, 49h
	int 21h				; free memory of former PSP and program location

.cantrelocate:			; We could display an error message here.
	or bp, _psprelocated		; do not relocate again if installing low (NC)
.return:
	pushf
	mov bx, cs
	add bx, 10h+transient_size_p
	mov ds, bx			; ds = handler segment
	popf
	retn

.mcberror:
	mov dx, msg.mcbcorrupted
.mcberrorcritical:
	jmp abort_msg

i23:
	mov word [ cs:$ ], (__JMP_REL8|__REL8__(.return-1)<<8)	; don't reenter
	call restorestate
.return:
	stc				; always abort program (what default DOS handler also does)
	retf

		; CHG:	-
restorestate:
	push ax
	push bx
	mov bx, word [ cs:restored.strategy ]
	mov ax, 5801h
	int 21h				; restore strategy

	mov bx, word [ cs:restored.umblink ]
	mov ax, 5803h
	int 21h				; restore UMB link
	pop bx
	pop ax
	retn

		; INP:	ds:si-> IISP interrupt entry
		;	al = interrupt number
		; OUT:	NC
		; CHG:	ah, es
hookinterrupt:
	push bx
	push dx
	mov ah, 35h
	int 21h			; get old handler
	mov [si +2], bx
	mov [si +4], es		; store

	mov ah, 25h
	mov dx, si		; ds:dx-> interrupt entry
	int 21h			; set new handler
	pop dx
	pop bx
	clc
	retn

onoff:
	mov dx, msg.off
	jz .off
	mov dx, msg.on
.off:				; fall through and display either
disp_msg:
	 push ds
	 push ax
	push cs
	pop ds
	mov ah, 09h
	int 21h
	 pop ax
	 pop ds
	retn


disp_al_hex:
		push cx
		mov cl, 4
		ror al, cl
		call .nibble
		ror al, cl
		pop cx
.nibble:
		push ax
		and al, 0Fh
		add al, '0'
		cmp al, '9'
		jbe .isdigit
		add al, 'A'-('9'+1)
.isdigit:
		xchg dl, al
		mov ah, 02h
		int 21h
		pop ax
		retn

	; Following call: Display number in ax decimal
	; all registers preserved except dx
disp_ax_dec:			; ax (no leading zeros)
	; In: number in ax
	; Out: displayed
		push bx
		xor bx, bx
.pushax:
		push ax
.pushend:
		or bl, bl
		jz .nobl
		sub bl, 5
		neg bl
.nobl:
		push cx
		mov cx, 10000
		call .divide_out
		mov cx, 1000
		call .divide_out
		mov cx, 100
		call .divide_out
		mov cl, 10
		call .divide_out
							; (Divisor 1 is useless)
		add al, '0'
		xchg dl, al
		mov ah, 02h
		int 21h
		pop cx
		pop ax
		pop bx					; Caller's register
		retn

.divide_out:
	; In: ax = number
	;     cx = divisor
	; Out: ax = remainder of operation
	;      result displayed
		push dx
		xor dx, dx
		div cx				; 0:ax / cx
		push dx				; remainder
		dec bl
		jnz .nobl2
		or bh, 1
.nobl2:
		or bh, al
		jz .leadingzero
		add al, '0'
		xchg dl, al
		mov ah, 02h
		int 21h				; display result
 .leadingzero:
		pop ax				; remainder
		pop dx
		retn

msg:
			; The version string is international
.name:		db "RxANSI ",RXANSI_VERP_STR,"beta3: ",36
%strcat _LANGFILE "RxANSIT.",_LANG
	%include _LANGFILE

restored:
.umblink:	dw 0
.strategy:	dw 0

parameters:			; Valid switches on the command line, followed
				; by an option selector byte with one bit set.
	db "A", _ansistate
	db "L", _lessstate
	db "J", _installlow
	db "P", _relocatepsp
	db "N", _installnew
	db "U", _uninstall
			; These undocumented options change relocation behaviour:
	db "E", _holdenvironment
	db "T", _forcetoprelocation
	db 0FFh			; table end marker

			; The help switches are not in this table (and neither in the option
			; selector) because on invalid switches, we'll show the help anyway.

	endarea transient


	section RESIDENT vstart=0 align=16
				; The handler is later copied into a seperate memory block.
				; The vstart= value tells NASM to use all addresses as if this
				; location is actually at vstart. (In this case, like if it's
				; at the start of the segment, not at 100h+start_size.)


%include "RXANSI.ASM"

%if _UPXPAD && (transient_size_p*16 + rxansi_handler_size) < 3062
		times 3062 -(transient_size_p*16 + rxansi_handler_size) db 90h
%endif